@@ -7,7 +7,7 @@ use strict;
 
 package XML::Compile::Cache;
 use vars '$VERSION';
-$VERSION = '0.97';
+$VERSION = '0.98';
 
 use base 'XML::Compile::Schema';
 
@@ -28,11 +28,11 @@ sub init($)
     $self->{XCC_wopts}  = delete $args->{opts_writers} || [];
     $self->{XCC_undecl} = delete $args->{allow_undeclared} || 0;
 
-    $self->{XCC_rcode}  = {};
+    $self->{XCC_rcode}  = {};  # compiled code refs
     $self->{XCC_wcode}  = {};
-    $self->{XCC_dropts} = {};
+    $self->{XCC_dropts} = {};  # declared opts
     $self->{XCC_dwopts} = {};
-    $self->{XCC_uropts} = {};
+    $self->{XCC_uropts} = {};  # undeclared opts
     $self->{XCC_uwopts} = {};
 
     $self->{XCC_readers} = {};
@@ -42,6 +42,8 @@ sub init($)
           = $self->_namespaceTable(delete $args->{prefixes});
     my %a = map { ($_->{prefix} => $_) } values %$p;
     $self->{XCC_prefixes} = keys %$p ? \%a : $p;
+    $self->typemap($args->{typemap});
+    $self->xsiType($args->{xsi_type});
 
     if(my $anyelem = $args->{any_element})
     {   # the "$self" in XCC_ropts would create a ref-cycle, causing a
@@ -115,6 +117,34 @@ sub prefixed($)
 }
 
 
+sub typemap(@)
+{   my $self = shift;
+    my $t    = $self->{XCC_typemap} ||= {};
+    my @d    = @_ > 1 ? @_ : !defined $_[0] ? ()
+             : ref $_[0] eq 'HASH' ? %{$_[0]} : @{$_[0]};
+    while(@d) { my $k = $self->findName(shift @d); $t->{$k} = shift @d }
+    $t;
+}
+
+
+sub xsiType(@)
+{   my $self = shift;
+    my $x    = $self->{XCC_xsi_type} ||= {};
+
+    my @d    = @_ > 1 ? @_ : !defined $_[0] ? ()
+             : ref $_[0] eq 'HASH' ? %{$_[0]} : @{$_[0]};
+
+    while(@d)
+    {   my $k = $self->findName(shift @d);
+        my $a = shift @d;
+        push @{$x->{$k}}, ref $a eq 'ARRAY' ? map ($self->findName($_), @$a)
+          : $self->findName($a);
+    }
+
+    $x;
+}
+
+
 sub allowUndeclared(;$)
 {   my $self = shift;
     @_ ? ($self->{XCC_undecl} = shift) : $self->{XCC_undecl};
@@ -242,7 +272,9 @@ sub mergeCompileOptions($$$)
       : ($self->{XCC_wopts}, $self->{XCC_dwopts}{$type});
 
     my %p    = %{$self->{XCC_namespaces}};
-    my %opts = (prefixes => \%p, hooks => [], typemap => {}, xsi_type => {});
+    my %t    = %{$self->{XCC_typemap}};
+    my %x    = %{$self->{XCC_xsi_type}};
+    my %opts = (prefixes => \%p, hooks => [], typemap => \%t, xsi_type => \%x);
 
     # flatten list of parameters
     my @take = map {!defined $_ ? () : ref $_ eq 'ARRAY' ? @$_ : %$_ }
@@ -263,16 +295,26 @@ sub mergeCompileOptions($$$)
         }
         elsif($opt eq 'typemap')
         {   $val ||= {};
-            @{$opts{typemap}}{keys %$val} = values %$val;
+            if(ref $val eq 'ARRAY')
+            {   while(@$val)
+                {   my $k = $self->findName(shift @$val); 
+                    $t{$k} = shift @$val;
+                }
+            }
+            else
+            {   while(my($k, $v) = each %$val)
+                {   $t{$self->findName($k)} = $v;
+                }
+            }
         }
         elsif($opt eq 'key_rewrite')
         {   unshift @{$opts{key_rewrite}}, ref $val eq 'ARRAY' ? @$val : $val;
         }
         elsif($opt eq 'xsi_type')
         {   while(my ($t, $a) = each %$val)
-            {   my @a = ref $a eq 'ARRAY' ? @$a : $a;
-                push @{$opts{xsi_type}{$self->findName($t)}}
-                   , map $self->findName($_), @a;
+            {   my @a = ref $a eq 'ARRAY' ? map($self->findName($_), @$a)
+                  : $self->findName($a);
+                push @{$x{$self->findName($t)}},  @a;
             }
         }
         else
@@ -297,11 +339,14 @@ sub _cleanup_hooks($)
     $hooks;
 }
 
+my %need = (READER => [1,0], WRITER => [0,1], RW => [1,1]);
+$need{READERS} = $need{READER};
+$need{WRITERS} = $need{WRITER};
+
 sub _need($)
-{    $_[1] eq 'READER' ? (1,0)
-   : $_[1] eq 'WRITER' ? (0,1)
-   : $_[1] eq 'RW'     ? (1,1)
-   : error __x"use READER, WRITER or RW, not {dir}", dir => $_[1];
+{   my $need = $need{$_[1]}
+       or error __x"use READER, WRITER or RW, not {dir}", dir => $_[1];
+    @$need;
 }
 
 # support prefixes on types
@@ -336,8 +381,6 @@ sub compileType($$@)
 
 #----------------------
 
-#----------------------
-
 
 sub declare($$@)
 {   my ($self, $need, $names, @opts) = @_;
